# Endangered Data Week, April 17-21, 2017
# UVa Library
# Web Scraping with R

# only submit these lines if you do not have the packages already installed
install.packages("rvest")
install.packages("magrittr")
install.packages("pbapply")
install.packages("readr")
install.packages("tidyr")
install.packages("dplyr")
install.packages("ggplot2")
install.packages("stringr")


library(rvest)
library(purrr) # for walk function
library(pbapply) # progress bar (pb) apply


# Example 1 ---------------------------------------------------------------

# set working directory; this is where you want to download the files;
# the below setwd() code works on my computer

# Use Session...Set Working Directory...Choose Directory in RStudio Menu to
# choose a folder on your computer.

setwd("C:/Users/jcf2d/Desktop/doe_files")

# https://www2.ed.gov/programs/osepidea/618-data/state-level-data-files/index.html

URL <- "https://www2.ed.gov/programs/osepidea/618-data/state-level-data-files/index.html"
page <- read_html(URL)
fileURL <- page %>% 
  html_nodes('a[href$=".csv"]') %>% 
  html_attr("href") %>% 
  paste0("https://www2.ed.gov",.)

# lapply(fileURL, function(x)download.file(x, destfile = basename(x)))
walk(fileURL[1:5], ~download.file(., destfile = basename(.)))



# Example 2 ---------------------------------------------------------------


# Web scraping bestbuy.com
# refrigerators with french doors
# get Description, Model, Price, average review

# go to http://www.bestbuy.com/ and search for "refrigerators with french doors"

# Resulting URL is long
# http://www.bestbuy.com/site/searchpage.jsp?st=refrigerators+with+french+doors&_dyncharset=UTF-8&id=pcat17071&type=page&sc=Global&cp=1&nrp=&sp=&qp=&list=n&af=true&iht=y&usc=All+Categories&ks=960&keys=keys

URL <- "http://www.bestbuy.com/site/searchpage.jsp?st=refrigerators+with+french+doors&_dyncharset=UTF-8&id=pcat17071&type=page&sc=Global&cp=1&nrp=&sp=&qp=&list=n&af=true&iht=y&usc=All+Categories&ks=960&keys=keys"
page <- read_html(URL)

# Now let's figure out how to select the information we want to scrape

# Description
# found ".sku-title a" with SelectorGadget
# Description contains Brand, size, description and color; 
# we can separate them out if we want after scraping
description <- page %>% html_nodes(".sku-title a") %>% html_text()

# pull Brand out of description; Brand appears to always preceded a dash.
# use str_extract() from the string package
brand <- stringr::str_extract(string = description, 
                              pattern = "[[:alpha:]]*(?= \\-)")


# Model number
# ".model-number .sku-value" found with SelectorGadget
model <- page %>% html_nodes(".model-number .sku-value") %>% html_text()


# Price
# Is there always a price? No! Sometimes it says "See price in cart"

# Looking at the source code we see there is an attribute called "data-price"
# inside div.list-item that contains the price if one is displayed.
page %>% html_nodes("div.list-item") %>% html_attr("data-price")

# The advantage to this approach is that we'll always get an NA if there is no 
# price. There is always a "div.list-item". We could also scrape the box
# containing the price. In that case we get the sale price and reg price, or the
# phrase "See price in cart" if there is no price. But that results in extra
# clean up.

price <- page %>% html_nodes("div.list-item") %>% 
  html_attr("data-price") %>% as.numeric()

# average review

# Again we use attributes. Inside div.list-item there are attributes
# "data-average-rating" and "data-review-count"
avg.rating <- page %>% html_nodes("div.list-item") %>% 
  html_attr("data-average-rating") %>% as.numeric()

review.count <- page %>% html_nodes("div.list-item") %>% 
  html_attr("data-review-count") %>% as.numeric()


# Combine into data frame
# set stringsAsFactors = FALSE, else text converted to Factor
dat <- data.frame(description, brand, model, price, avg.rating, review.count, 
                  stringsAsFactors = FALSE)

# That's just page 1 results. Want to get all results. 

# Turn what we did into a function called getData

# Give this function a URL from BestBuy search results and it should return a
# data frame
getData <- function(url){
  page <- read_html(url)
  description <- page %>% html_nodes(".sku-title a") %>% html_text()
  brand <- stringr::str_extract(string = description, 
                                pattern = "[[:alpha:]]*(?= \\-)")
  model <- page %>% html_nodes(".model-number .sku-value") %>% html_text()
  price <- page %>% html_nodes("div.list-item") %>% 
    html_attr("data-price") %>% as.numeric()
  avg.rating <- page %>% html_nodes("div.list-item") %>% 
    html_attr("data-average-rating") %>% as.numeric()
  review.count <- page %>% html_nodes("div.list-item") %>% 
    html_attr("data-review-count") %>% as.numeric()
  data.frame(description, brand, model, price, avg.rating, 
             review.count, stringsAsFactors = FALSE)
}

# Try it out
getData(URL)

# Now let's apply our getData function to a vector of URLs (ie, all pages of
# results). We need to get URLs

# notice cp=1, cp=2, cp=3, in the URLs.
# They're the only part that changes as we page through results

# Therefore we can manually build a vector of URLs %>% 
allURLs <- paste0("http://www.bestbuy.com/site/searchpage.jsp?cp=",
                  1:5,
                  "&searchType=search&st=refrigerators%20with%20french%20doors&_dyncharset=UTF-8&id=pcat17071&type=page&sc=Global&nrp=&sp=&qp=&list=n&af=true&iht=y&usc=All%20Categories&ks=960&keys=keys")

# Now "lapply" our function to each URL in our vector.

# We'll use the pbapply function. The "pb" stands for progress bar and allows us
# to see the progress.
allDat <- pblapply(allURLs, getData)

# The result is a list. We can collapse into one data frame using the dplyr
# function bind_rows():
allDatDF <- dplyr::bind_rows(allDat)

# This works for the one search we performed. But what if we wanted to 
# generalize this code so we could search BestBuy for any product and get a data
# frame of results, all within R?

# We need to be able to do two things:
# 1. submit the search form
# 2. automatically assemble a list of URLs for our getData function

# Let's start on #2 first.

# We'll create a function to generate a vector of URLs for whatever we search for.

# Notice there are 24 results per page and the total number of pages is not 
# always returned. Try searching "TV", for example. However the total number of 
# items is listed next to All Items. Therefore we could divide that number by 24
# and round to next largest integer if we get a decimal.

# use SelectorGadget to select the number next to All Items
# The result is ".active .count"
page %>% html_nodes(".active .count") %>% html_text()

# Use the parse_number() function from the readr package to handle numbers with
# commas (eg: 5,664) and to remove parentheses:
total <- page %>% html_nodes(".active .count") %>% 
  html_text() %>% readr::parse_number()


# divide by 24 and use ceiling to get number of pages, and then generate sequence
pages <- ceiling(total/24) %>% seq()

# create vector of URLs using pages instead of 1:5
allURLs <- paste0("http://www.bestbuy.com/site/searchpage.jsp?st=refrigerators+with+french+doors&_dyncharset=UTF-8&id=pcat17071&type=page&sc=Global&cp=",
                  pages,
                  "&nrp=&sp=&qp=&list=n&af=true&iht=y&usc=All+Categories&ks=960&keys=keys")


# To generalize the function, we need to get the URL produced by the search.

# Let's see what we get when we submit the form using submit_form()

# Recall we need to use html_session instead of read_html when we intend to
# submit a form:
bb <- html_session("http://www.bestbuy.com/")

# to see form fields use html_form()
bb %>% html_form()

# Notice  <input text> 'st':
# We want to set values for that form field

# save form
f0 <- html_form(bb)[[1]]
# set values using set_values()
f1 <- set_values(form = f0, st = "refrigerators with french doors")
# and submit the form with our values and save
page <- bb %>% submit_form(f1)

# the result is the same as before plus more; page is a parsed html file that we
# can use html_nodes on:
page %>% html_nodes(".active .count")
page %>% html_nodes("div.list-item") %>% 
  html_attr("data-review-count")

# but it also has the URL generated from the search
page$url

# We can use page$url to build our vector of search result URLs

# Notice we can use strsplit to split the URL at "cp="
strsplit(page$url, "cp=1")
unlist(strsplit(page$url, "cp=1"))
parts <- unlist(strsplit(page$url, "cp=1"))
cp <- paste0("cp=",pages)
paste0(parts[1],cp,parts[2])

# Let's use all this to build a master function to search and scrape Best Buy
bbSearch <- function(x){
  bb <- html_session("http://www.bestbuy.com/")
  f0 <- html_form(bb)[[1]]
  f1 <- set_values(form = f0, st = x)
  page <- bb %>% submit_form(f1)
  # create vector of URLS 
  total <- page %>% html_nodes(".active .count") %>% 
    html_text() %>% readr::parse_number()
  pages <- ceiling(total/24) %>% seq()
  cp <- paste0("cp=",pages)
  parts <- unlist(strsplit(page$url, "cp=1"))
  allURLs <- paste0(parts[1],cp,parts[2])
  # scrape site using getData
  allDat <- pblapply(allURLs, getData)
  dplyr::bind_rows(allDat)
}

# test it
fridgeData <- bbSearch("refrigerators with french doors")

# Try for "TV"; does everything work?
tvData <- bbSearch("TV")


# some basic analysis by brand


# Break down of brands
table(fridgeData$brand)

# mean avg.rating by brand
library(dplyr)
fridgeData %>% 
  group_by(brand) %>% 
  summarize(score = mean(avg.rating, na.rm=TRUE),
            reviews = sum(review.count))

# Top 10 highest rated models for those with at least 100 reviews
fridgeData %>% 
  arrange(desc(avg.rating)) %>% 
  filter(review.count > 100) %>% 
  head(n=10) %>% 
  select(description, avg.rating, review.count, price) %>% 
  View()


# distribution of prices by brand
library(ggplot2)
ggplot(fridgeData, aes(x = brand, y = price)) + geom_boxplot()



